#!/usr/bin/perl -w

#
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2012.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#
##############################################################################

use warnings;
use strict;
use Time::Local;


# To enable this script to kill abandoned cqrpc processes, change to 1.  By default, simply report abandoned processes (don't kill).
my $enable_kill = 0;

# To disable this script on Unix, change the value of 'enable_for_unix' to 0.  By default, it is enabled for Unix.
my $enable_for_unix = 1;

# To enable this script on Windows, change the value of 'enable_for_unix' to 1.  By default, it is disabled for Windows.
my $enable_for_windows = 0;


sub Usage
{
	print <<EOF;

This PERL script is used to find abandoned cqrpc processes and terminate them.

Usage: cqperl cqrpccleanup.pl [-lifetime <minutes>]

Parameters:
  -lifetime  How long a process must be running before it is considered to 
             be terminated.

This PERL script is used by CM server to clean up any abandoned cqrpc processes.
It is invoked by CM server whenever a recycle event condition occurs.

EOF
}


my $DEBUG = $ENV{CQ_DEBUG_CQRPCCLEANUP} || 0;
my $PROCESS_NAME = "cqrpc";
my $IS_WINDOWS = ($^O =~ /MSWin32/);


# Define default values...
my(%ARG) = (
				LIFETIME => '-1'
	);

# See if they requested usage info...
if ((@ARGV != 2) || ($ARGV[0] =~ /^-[h\?]/i))
{
	Usage();
	exit 0;
}

# Parse the user's command-line parameters...
my $idx = 0;
my $arg = $ARGV[$idx++];

if ($arg =~ /-lifetime/i)
{
	$ARG{LIFETIME} = $ARGV[$idx++];
}
else
{
	Usage();
	exit 1;
}


if (($enable_for_unix and ! $IS_WINDOWS) || ($enable_for_windows and $IS_WINDOWS))
{
	# Call the function to look for abandoned cqrpc processes and terminate them
	findAndKillAbandonedProcess($PROCESS_NAME, $ARG{LIFETIME});
}
else
{
	print "Abandoned '$PROCESS_NAME' cleanup is disabled.\n";
}

exit 0; # This assumes all is well, but that's not always the case


##############################################################################
# Iterates over list of processes and kills anything that's still active 
# (based on 'killTimeLimit') for longer than the allowed time
sub findAndKillAbandonedProcess
{
	my ($processName, $killTimeLimit) = @_;

	# Whatever the CM server recycle lifetime limit is, add to it 10%.
	# This is to give some window before we attempt to terminate a process
	$killTimeLimit = int(($killTimeLimit * 110) / 100);	# add 10% as a safeguard

	print ">>>> killTimeLimit[$killTimeLimit]\n" if $DEBUG;

	print "Looking for abandoned '$processName' processes with up-time over '$killTimeLimit' minutes\n";

	my $totalKilled = 0;

	if ($IS_WINDOWS)
	{
		$totalKilled = findAndKillAbandonedProcess_Win($processName, $killTimeLimit);
	}
	else
	{
		$totalKilled = findAndKillAbandonedProcess_Unix($processName, $killTimeLimit);
	}

	print "Total abandoned '$processName' processes found (and may have been terminated): $totalKilled\n";
}


##############################################################################
# Unix version
sub findAndKillAbandonedProcess_Unix
{
	my ($processName, $killTimeLimit) = @_;

	my $totalKilled = 0;

	my @processes_list = `ps -eo pid,etime,comm,pmem,args 2>&1 | grep $processName 2>&1` or warn "Error whie executing 'ps': $!\n";
	if ($? > 0) # Check for the exit code of 'ps'
	{
		print "Exit error code: $?\n";
		return $totalKilled;
	}

	my $found_abandoned_process = 0;
	my $error_msg = "";

	# Iterate over the result of 'ps' looking for abandoned processes to kill
	foreach my $process_line (@processes_list)
	{
		$found_abandoned_process = 0;
		$error_msg = "";
		
		$process_line = trim($process_line);

		print ">>>> process_line[$process_line]\n" if $DEBUG;
		
		# This line may be a candidate for our process
		if ($process_line !~ m/$processName/i)
		{
			next;
		}
		
		# Split on spaces; all that we care about is the 'pid' and the 'etime'
		# exp: 10683  2-01:48:07 xprop            0.1 /usr/bin/xprop -id 0x102 -spy
		my @ps_items = split(/ +/, $process_line);
		
		if ($DEBUG)
		{
			print ">>>>";
			foreach my $item (@ps_items)
			{
				print "[".$item."] ";
			}
			print "\n";
		}
		
		if (scalar(@ps_items) < 5) # Make sure we got at least 5 items
		{
			$error_msg = "Unexpected line format (less than 5 items returned): [$process_line]";
		}
		
		if ($error_msg eq "")
		{
			# Check to see if this is exactly our process
			if ($ps_items[2] ne $processName)
			{
				next;
			}
			
			# No issues so far, find out if this process has been running longer than the allowed time
			my $pid = $ps_items[0];
			my $etime = $ps_items[1];
			
			my $etime_min;
			
			($etime_min, $error_msg) = getEtimeAsMinutes($etime);
			
			print ">>>> etime_min:[$etime_min] error_msg:[$error_msg]\n" if $DEBUG;
			
			# If we calculated the etime without any issues, keep going
			if ($error_msg eq "")
			{
				print "Found a '$processName' process: $process_line \n";

				# All is well, now if this 'pid' is past it's recycle time, kill it
				if ($etime_min >= $killTimeLimit)
				{
					print "Found an abandoned process: [$pid]\n";

					$found_abandoned_process = 1;

					if ($enable_kill == 1)
					{
						print "Terminating an abandoned process: [$pid]\n";

						`kill -9 $pid`;
						if ($? > 0) # Check for the exit code of 'kill'
						{
							print "Tertmination of abandoned process failed with exit code of: $?\n";
						}
						else
						{
							$totalKilled++;
						}
					}
					else
					{
						print "Skipping termination of abandoned process because termination is disabled.\n";
					}
				}
				else
				{
					print "Not an abandoned process: [$pid]\n";
				}
			}
		}

		# Report any error
		if ($error_msg ne "")
		{
			print "Error: $error_msg\n";
		}

		# If an abandoned process is found (or an error was encountered)
		# and we are not able to kill it, report it
		if (($found_abandoned_process == 1) and ($error_msg ne ""))
		{
			print "Failed to terminate abandoned process due to error.\n";
		}
	}

	return $totalKilled;
}


##############################################################################
# Windows version
sub findAndKillAbandonedProcess_Win
{
	my ($processName, $killTimeLimit) = @_;

	my $totalKilled = 0;
	
	my $processes_list = `WMIC PROCESS GET NAME, CREATIONDATE, PROCESSID 2>&1` or warn "Error while executing 'wmic': $!\n";
	if ($? > 0) # Check for the exit code of 'wmic'
	{
		print "Exit error code: $?\n";
		return $totalKilled;
	}

	my $found_abandoned_process = 0;

	my @lines = split("\n", $processes_list);
	for (@lines)
	{
		print ">>>> lines[$_]\n" if $DEBUG;

		if ($_ =~ /$processName/i) 
		{
			print "Found a '$processName' process: $_ \n";

			# Split on spaces; all that we care about is the 'CreationDate' and the 'ProcessId'
			# exp: 20120828175514.524554-240  cqrpc.exe                 23136
			my @processstrings = split(/ +/, $_);
			my $pid = $processstrings[2];

			# Extract the year, month, day, hour, minutes, seconds from the date/time string.
			# exp: 20120828175514 ~ August 28 2012, 17:55:14
			my ($sec) 	= substr $processstrings[0], 12, 2;
			my ($min) 	= substr $processstrings[0], 10, 2;
			my ($hour)	= substr $processstrings[0], 8, 2;
			my ($mday)	= substr $processstrings[0], 6, 2;
			my ($mon)  	= substr $processstrings[0], 4, 2;
			my ($year)	= substr $processstrings[0], 0, 4;

			# Convert the extracted values to a time that can be differentiated later.
			my $processStartTime = timelocal( $sec, $min, $hour, $mday, $mon-1, $year - 1900 );
			# Get the current time
			my $currentDateTime = time();

			# Calculate time difference from now and the process start time in seconds.
			my $upTimeSec = $currentDateTime - $processStartTime;
			# Convert the time difference to minutes.
			my $upTimeMin = $upTimeSec / 60;

			print ">>>> upTimeMin[$upTimeMin]\n" if $DEBUG;

			my $kill_msg = "";

			# If this 'pid' is past it's recycle time, kill it
			if ($upTimeMin >= $killTimeLimit)
			{
				print "Found an abandoned process: [$pid]\n";

				$found_abandoned_process = 1;

				if ($enable_kill == 1)
				{
					print "Terminating an abandoned process: [$pid]\n";

					$kill_msg = `taskkill /PID $pid /F /T`;
					if ($? > 0) # Check for the exit code of 'taskkill'
					{
						print "Termination of abandoned process failed with exit code of: $?\n";
					}
					else
					{
						$totalKilled++;
					}

					# Report any error
					if (defined($kill_msg) && ($kill_msg !~ /"SUCCESS"/i))
					{
						print "Error: $kill_msg\n";
					}
				}
				else
				{
					print "Skipping termination of abandoned process because termination is disabled.\n";
				}
			}
			else
			{
				print "Not an abandoned process: [$pid]\n";
			}

			# If an abandoned process is found (or an error was encountered)
			# and we are not able to kill it, report it
			if (($found_abandoned_process == 1) and ($kill_msg !~ /"SUCCESS"/i))
			{
				print "Failed to terminate abandoned process.\n";
			}
		}
	}

	return $totalKilled;
}


##############################################################################
# Given 'etime' based on format per: 
#    http://publib.boulder.ibm.com/infocenter/pseries/v5r3/index.jsp?topic=/com.ibm.aix.cmds/doc/aixcmds4/ps.htm
# return it as minutes
# Format of 'etime' is: [[ dd -] hh: ]mm :ss
#
# Returns two values: elaps time in minutes and an error if any
sub getEtimeAsMinutes
{
	my ($etime) = @_;
	
	my @days_part = ();
	my @time_part = ();
	
	my $days = 0;
	my $hours = 0;
	my $minutes = 0;
	my $seconds = 0;
	
	my $total_min = -1;
	my $error_msg = "";
	
	@days_part = split(/-/, $etime);
	
	# Deal with date / time part
	if (scalar(@days_part) == 2)
	{
		# We have a day part (and time part)
		# exp: 2-01:48:07
		$days = $days_part[0]; # This is our days count
		@time_part = split(/:/, $days_part[1]); # this has our time part
	}
	elsif (scalar(@days_part) == 1)
	{
		# Only time part
		# exp: 01:48:07 OR 48:07
		$days = 0;
		@time_part = split(/:/, $days_part[0]); # this has our time part
	}
	else
	{
		# We have an error
		$error_msg = "More than one '-' was found in '$etime'";
	}
	
	if ($error_msg eq "")
	{
		# Deal with time part which maybe as 01:48:07 OR 48:07
		if (scalar(@time_part) == 3)
		{
			$hours = $time_part[0];
			$minutes = $time_part[1];
			$seconds = $time_part[2];
		}
		elsif (scalar(@time_part) == 2)
		{
			$hours = 0;
			$minutes = $time_part[0];
			$seconds = $time_part[1];
		}
		else
		{
			# We have an error
			$error_msg = "The wrong number of ':' was found in '$etime'";
		}
		
		if ($error_msg eq "")
		{
			$days = stripLeadingZero($days);
			$hours = stripLeadingZero($hours);
			$minutes = stripLeadingZero($minutes);
			$seconds = stripLeadingZero($seconds);	# we don't use seconds
			
			print ">>>> days:[$days], hours:[$hours], minutes:[$minutes], seconds:[$seconds]\n" if $DEBUG;
			
			# Convert to minutes
			$total_min = 0;
			
			$total_min += $days * (24 * 60);
			$total_min += $hours * 60;
			$total_min += $minutes;
		}
	}
	
	return ($total_min, $error_msg);
}


##############################################################################
# Removes white characters from start and end of a string
sub trim
{
	my ($string) = @_;
	
	if (defined($string))
	{
		$string =~ s/^\s+//;
		$string =~ s/\s+$//;
	}
	
	return $string;
}


##############################################################################
# Removes leading '0' from the string (only the first one gets removed)
sub stripLeadingZero
{
	my ($string) = @_;
	
	if (defined($string) and length($string) > 1)
	{
		$string =~ s/^0//;
	}
	
	return $string;
}
